from .base_model import Model
    
class Llama(Model):
    """
    Wrapper around all Llama based models.
    """
    ...
    
    
class GptJ(Model):
    """
    Wrapper around all GPTJ based models.
    """
    ...
    
class Gpt2(Model):
    """
    Wrapper around all GPT2 based models.
    """
    ...
    
class Bloom(Model):
    """
    Wrapper around all Bloom based models.
    """
    ...
    
class GptNeoX(Model):
    """
    Wrapper around all GPT-NeoX based models.
    """
    ...

class Mpt(Model):
    """
    Wrapper around all Mpt based models.
    """
    ...